# affilmat.py -- routines for analysis of affiliation matrices

# Copyright (c) 2012 Claude Rubinson

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <http://www.gnu.org/licenses/>.

from numpy import *
from pptable import *

class AffilMatError(Exception):
    """Base class for exceptions in affilmat module."""
    pass

class AffilMatConstructionError(AffilMatError):
    """Exception raised when failing to construct an affiliation matrix."""
    pass

class AffilMat(object):
    """Affiliation matrix class."""

    def __init__(self, inmatrix, actors, events):

        try:
            inmatrix = mat(inmatrix)
        except ValueError, exc:
            raise AffilMatConstructionError, "Malformed input data; cannot coerce to matrix"

        #
        # validation
        #

        # number of actors must equal number of rows
        if len(actors) != inmatrix.shape[0]:
            raise AffilMatConstructionError, "Number of actors does not equal number of rows"
        # number of events must equal number of colums
        if len(events) != inmatrix.shape[1]:
            raise AffilMatConstructionError, "Number of events does not equal number of cols"

        # input data can be 0 or 1
        for i in inmatrix.flat:
            if i not in (0,1):
                raise AffilMatConstructionError, ("Invalid value in input: %s" % i)

        # instantiation
        self.matrix = inmatrix
        self.actors = actors
        self.events = events
        return None

    def asmatrix(self):
        return self.matrix

    def __repr__(self):
        return str(self.matrix)

    def __str__(self):
        tbl = []
        tbl.append([' '] + list(self.events))
        for i,row in enumerate(self.matrix.A):
            tbl.append([self.actors[i]] + str(row).strip('[]').split())
        return pp(tbl)

    def _uniqify(self, dimension='actors', delim=','):
        """Identify and combine identical actors or events.

        Set dimension= to 'actors' to uniqify actors and to 'events'
        to uniqify events.
        
        This method is for internal use and shouldn't be called
        directly.  Instead, call AffilMat.uniqify()
        """

        if dimension.lower() not in ('actors','events'):
            raise ValueError, 'Invalid dimension: %s' % dimension

        seen_rows = []
        seen = []

        if dimension == 'actors':
            working = self.matrix
            labels = self.actors
        else:
            working = transpose(self.matrix)
            labels = self.events

        for i, row in enumerate(working.A):
            lrow = list(row)
            if lrow in seen_rows:
                initial_row = seen_rows.index(lrow)
                seen[initial_row] = seen[initial_row] + delim + labels[i]
            else:
                seen_rows.append(lrow)
                seen.append(labels[i])

        if dimension == 'actors':
            uniqified = AffilMat(mat(seen_rows), seen, self.events)
        else:
            uniqified = AffilMat(transpose(mat(seen_rows)), self.actors, seen)

        return uniqified

    def uniqify(self, dimension=None, delim=','):
        """Identify and combine identical actors and/or events.

        By default, uniqify both actors and events.  Set dimension= to
        'actors' to uniqify only actors and to 'events' to uniqify
        only events."""

        if not dimension:
            return self._uniqify(dimension='actors', delim=delim)._uniqify(dimension='events', delim=delim)
        elif dimension.lower() == 'actors':
            return self._uniqify(dimension='actors', delim=delim)
        elif dimension.lower() == 'events':
            return self._uniqify(dimension='events', delim=delim)
        else:
            raise ValueError, 'Invalid dimension: %s' % dimension

    def galois(self, delim=','):
        """Construct Galois matrix.

        Galois lattices are described in Wasserman and Faust (1994),
        section 8.6.1.  This method generates a galois matrix, which
        can be used to produce a galois lattice, in the form of
        dictionary.  The dictionary's keys are the rows of the matrix
        and the values are two-element tuples of the corresponding
        actor and event label.

        The process of constructing a Galois matix is described in
        Breiger (2000) "A toolkit for practice theory," Poetics (27),
        pages 102-104.  In brief: we first uniqify the matrix and then
        take the intersection of each pair of rows.  These
        intersections form the galois matrix.  Duplicate rows are
        eliminated.  Rows representing the Universal and Empty sets
        are added, if they don't exist.  This ensure that the lattice
        will close.

        Rows are labeled with the actors and events to which they
        correspond.  Actors are affixed to their corresponding highest
        order row, while events are affixed to their corresponding
        lowest order row.  This is not configurable.  If you wish flip
        the matrix, you'll need to transpose your affiliation matrix's
        actors and events.
        """
        
        # take intersections of all pairs of rows
        intersections = set()
        inmatrix = self.uniqify(delim=delim)
        for row1 in inmatrix.asmatrix().A:
            for row2 in inmatrix.asmatrix().A:
                intersections.add(tuple(row1 & row2))
        # append null and universal rows if they don't already exist
        intersections.add(tuple([0] * len(inmatrix.events)))
        intersections.add(tuple([1] * len(inmatrix.events)))

        # attach labels to galois matrix
        galois_mat = []
        galois_actors = {}
        galois_events = {}
        
        # actors are affixed at their corresponding highest order row.
        # (galois_actors() is guaranteed include an item for every
        # row)
        for intersection in intersections:
            galois_mat.append(intersection)
            try:
                index = inmatrix.matrix.tolist().index(list(intersection))
                actor = inmatrix.actors[index]
            except ValueError:
                if not any(intersection):  # Null set
                    actor = 'Null'
                elif all(intersection):  # Universal set
                    actor = 'Universal'
                else:
                    actor = None
            galois_actors[intersection] = actor

        # events are affixed to their corresponding lowest order row.
        # For each event, get all corresponding rows from galois
        # matrix and find the lowest order one.  (galois_events{} will
        # only include rows with a corresponding event)
        for i, event in enumerate(inmatrix.events):
            lowest_order=len(inmatrix.events)
            for row in galois_mat:
                if row[i]:
                    if sum(row) < lowest_order:
                        lowest_order = sum(row)
                        lowest_order_row = galois_mat.index(row)
            galois_events[galois_mat[lowest_order_row]] = event

        # construct galois matrix
        galois = {}
        for row in galois_mat:
            try:
                galois[row] = galois_actors[row], galois_events[row]
            except KeyError:  # raised if row has no corresponding event
                galois[row] = galois_actors[row], None

        return galois
        
if __name__ == "__main__":

    parties_mat = [[1,0,1],
                   [0,1,0],
                   [0,1,1],
                   [0,0,1],
                   [1,1,1],
                   [1,1,0]]
    parties_actors = ['Allison','Drew','Eliot','Keith','Ross','Sarah']
    parties_events = ['Party 1','Party 2','Party 3']
    parties = AffilMat(parties_mat, parties_actors, parties_events)

    print parties
    print
    print parties.galois()

    print; print

    breiger_mat = ('1,0,1,1,1; 1,0,1,1,1; 1,0,0,1,0; 0,0,1,0,1; 1,1,0,1,0; 0,1,0,1,0; 0,1,0,0,1; 0,0,1,0,1')
    breiger_actors = ('Rehnquist', 'Scalia', 'Kennedy', "O'Connor", "Stevens", "Blackmun", "Souter", "White")
    breiger_events = ('Crime', 'Economy', 'Civil Rights', 'First Amend', 'Due Process')

    breiger = AffilMat(breiger_mat, breiger_actors, breiger_events)

    print breiger
    print
    print breiger.galois()
